#!/bin/bash

#
#   Script for automated comparison of two struture files for NCC project
#
#   Produces summary of chemical fingerprint similarity (CFPs and ECFPs)
#   scores and counts of novel ring systems and Bemis-Murcko frameworks.
#   Deliberately high-level summary only to enable data-sharing.
#
#   Copyright 2014 NQuiX Ltd
#
#   Licensed under the Apache License, Version 2.0 (the 'License');
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the license at:
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

THISPROG=`basename $0`
MYPATH=`pwd`

#
#   Set default values for structural files and their identifier tags.
#

NCCROOT="NCC"
NCCID="csid"
COLROOT="test"
COLID="name"

#
#   Set default values for output file name, number of CPUs for parallel
#   calculations of ECFP similarity and number of structures per slice
#   in file splitting.
#

OUTFI="Analysis-Summary.txt"
NCPUS=4
SLICESIZE=100

#
#   Set file names for CFP generation and structure standardization parameters.
#   The names of the files are hardcoded to ensure consistency between groups.
#   The cfp.xml file is taken from the ChemAxon website.
#   The stanrep.xml file was generated with the ChemAxon Standardizer Editor tool.
#   Any queries regarding these files should be referred to ChemAxon.
#

FPRINT="cfp.xml"
STAND="stanrep.xml"

#
#   Substructural and property .txt files should be created and edited to conform
#   with in-house standards for structure filtering.
#   
#   For example, the PROPS file "Prop-Lip4.txt" should contain an expression such as:
#      (mass() < 500) && (logP() <= 5) && (donorCount() <= 5) && (acceptorCount() <= 10)
#   to provide Lipinski-style property filtering
#
#   whilst the SSQRY file "SS-Filters.txt" should contain a series of SMARTS strings for
#   substructures to be removed.
#

PROPS="Prop-Lip4.txt"
SSQRY="SS-Filters.txt"

#
#   Process command line.
#

while getopts ":c:e:hi:j:n:o:p:q:" opt; do
    case $opt in
	c)
	    NCCROOT=$OPTARG
	    NCCROOT=`echo $NCCROOT | sed 's|\.smi$||'`
	    NCCROOT=`echo $NCCROOT | sed 's|\.sdf$||'`
	    ;;
	e)
	    COLROOT=$OPTARG
	    COLROOT=`echo $COLROOT | sed 's|\.smi$||'`
	    COLROOT=`echo $COLROOT | sed 's|\.sdf$||'`
	    ;;
	h)
	    echo "$THISPROG" >&2
	    echo "" >&2
	    echo "-c comparison set of molecules for analysis (.sdf or .smi) [e.g. NCC set]" >&2
	    echo "   Filename root - a .smi or .sdf suffix will be removed if required" >&2
	    echo "-e existing collection of molecules (.sdf or .smi) [e.g. screening collection]" >&2
	    echo "   Filename root - a .smi or .sdf suffix will be removed if required" >&2
	    echo "-i comparison set identifier tag" >&2
	    echo "-j existing set identifier tag" >&2
	    echo "-n number of CPUs (crude parallelization for ECFP comparisons)" >&2
	    echo "-o output file" >&2
	    echo "-p property filters file (text file, SMARTS strings)" >&2
	    echo "-q substrutural query filters file (text file, CXN chemical terms)" >&2
	    echo "" >&2
	    echo "Example Usage (default values)" >&2
	    echo "$THISPROG -b $NCCROOT -i $NCCID -c $COLROOT -j $COLID -n $NCPUS -o $OUTFI -p $PROPS -q $SSQRY" >&2
	    echo "" >&2
	    echo "Configuration files for fingerprint generateion ($FPRINT) and" >&2
	    echo "structure standardization ($STAND) are deliberatly hardcoded - please don't change" >&2
	    echo "" >&2
	    echo "Note: Substructural and property filters are only applied to the comparison set" >&2
	    echo "      The existing collection should already be filtered" >&2
	    exit 0
	    ;;
	i)
	    NCCID=$OPTARG
	    ;;
	j)
	    COLID=$OPTARG
	    ;;
	n)
	    NCPUS=$OPTARG
	    ;;
	o)
	    OUTFI=$OPTARG
	    ;;
	p)
	    PROPS=$OPTARG
	    ;;
	q)
	    SSQRY=$OPTARG
	    ;;
	s)
	    STAND=$OPTARG
	    ;;
	\?)
	    echo "Invalid option: -$OPTARG" >&2
	    exit 1
	    ;;
	:)
	    echo "Option -$OPTARG requires an argument" >&2
	    exit 1
	    ;;
    esac
done

echo "Running $THISPROG -c $NCCROOT -i $NCCID -e $COLROOT -j $COLID -n $NCPUS -o $OUTFI -p $PROPS -q $SSQRY with $STAND and $FPRINT" >&2

#
#   Comparison Set: Convert SDF to SMILES (if required), Standardize Structures,
#   Apply Substructural and Property Filters and Generate FPs.
#
#   Note: Combined substructure and property filter via single jcsearch call
#   appears not to work. Referred to ChemAxon.
#

if [ $NCCROOT.sdf -nt $NCCROOT.smi ]; then
    echo "Converting $NCCROOT.sdf to $NCCROOT.smi"
    molconvert smiles:q -T $NCCID -Y -g -o $NCCROOT.smi $NCCROOT.sdf 2> $NCCROOT.molcon-err.log
fi

if [ -e $NCCROOT.smi ]; then
    echo "Standardizing $NCCROOT"
    standardize -c $STAND -g -f smiles:u,T$NCCID $NCCROOT.smi 2> $NCCROOT.stan-err.log | sort -u -k 1,1 | grep -v "^\s" | jcsearch -q $SSQRY -n --and -g -f smiles:q,T$NCCID | jcsearch -q "[*]" -e $PROPS -f smiles:q,T$NCCID -o $NCCROOT.std
else
    echo "File does not exist: $NCCROOT.smi"
    exit
fi

echo "Generating CFPs for $NCCROOT"
generatemd c $NCCROOT.std -k CF -c $FPRINT -D -o $NCCROOT.cfp

#
#   Existing Set: Convert SDF to SMILES (if required), Ensure Equivalent
#   Structure Standardization, No Filtering, Generate FPs.
#

if [ $COLROOT.sdf -nt $COLROOT.smi ]; then
    echo "Converting $COLROOT.sdf to $COLROOT.smi"
    molconvert smiles:q -T $COLID -Y -g -o $COLROOT.smi $COLROOT.sdf 2> $COLROOT.molcon-err.log
fi

if [ -e $COLROOT.smi ]; then
    echo "Standardizing $COLROOT"
    standardize -c stanrep.xml -g -f smiles:u,T$COLID $COLROOT.smi 2> $COLROOT.stan-err.log | sort -u -k 1,1 | grep -v "^\s" > $COLROOT.std
else
    echo "File does not exist: $COLROOT.smi"
    exit
fi

echo "Generating CFPs for $COLROOT"
generatemd c $COLROOT.std -k CF -c $FPRINT -D -o $COLROOT.cfp

#
#   Chemical Fingerprints: Calculate similarity score distribution and
#   proportion less then 40% similar to existing set.
#
#   Note - CXN software reports diversity rather than similarity.
#   Less than 40% similarity = >0.6 diversity.
#
#   Note - CFP scores can be calculated quickly via compr. Procedure
#   appears not to be valid for ECFPs even if decimal representation
#   is converted to binary. Go via screenmd instead. Small perl script
#   (screenmd-compress.pl) to reformat output and avoid very large
#   files - just capture closest compound, not all scores.
#

echo "Performing CFP comparison for $COLROOT versus $NCCROOT"
compr -Xmx12000m -f 1024 -g -i $COLROOT.cfp $NCCROOT.cfp -o $NCCROOT.$COLROOT.cfp.out
echo "CFP score distribution (of filter passes)" > $OUTFI
grep -v "minD" $NCCROOT.$COLROOT.cfp.out | awk '{printf "%.4f\n", 1-$2}' | sort -nr >> $OUTFI
echo "CFP count of compounds <40% similarity" >> $OUTFI
grep -v "id" $NCCROOT.$COLROOT.cfp.out | awk '{if ($2 > 0.6) { print $0 }}' | wc -l >> $OUTFI

echo "Generating ECFPs for $NCCROOT"
generatemd c $NCCROOT.std -k ECFP -o $NCCROOT.ecfp

echo "Performing ECFP comparison for $COLROOT versus $NCCROOT (may take some time ...)"
grep -v "\#SMILES" $COLROOT.std | split -a 5 -d -l $SLICESIZE - ECFPSLICE-
NSLICE=`ls ECFPSLICE-* | wc -l`
let NPERCPU=($NSLICE/$NCPUS)+1
echo "Processing $NSLICE slices of $SLICESIZE compounds across $NCPUS CPUs"
ls ECFPSLICE-* | split -l $NPERCPU

pidArr=()
for SPFILE in x[a-z][a-z]; do
    awk '{print "screenmd -Xmx2000m "$1" -k $NCCROOT.ecfp -g 1 -e 4 | perl $MYPATH/screenmd-compress.pl "$1" 2> /dev/null > "$1".ecfp.out"}' $SPFILE | sed 's/\.pl ECFPSLICE-\([0-9][0-9][0-9][0-9][0-9]\)/\.pl \1 $SLICESIZE/' > $SPFILE.sh
    chmod 755 $SPFILE.sh
    source $SPFILE.sh &
    pidArr+=($!)
done

wait ${pidArr[@]}

#   Combine similarity scores across data chunks and delete temporary files
perl screenmd-combine.pl ECFPSLICE-[0-9][0-9][0-9][0-9][0-9].ecfp.out | sort -n -k 1 > $NCCROOT.$COLROOT.ecfp.out
rm ECFPSLICE-[0-9][0-9][0-9][0-9][0-9]
rm ECFPSLICE-[0-9][0-9][0-9][0-9][0-9].ecfp.out
rm x[a-z][a-z]
rm x[a-z][a-z].sh

echo "ECFP score distribution (of filter passes)" >> $OUTFI
grep -v "Tanimoto" $NCCROOT.$COLROOT.ecfp.out | awk '{printf "%.4f\n", 1-$2}' | sort -nr >> $OUTFI
echo "ECFP count of compounds <40% similarity" >> $OUTFI
grep -v "ID" $NCCROOT.$COLROOT.ecfp.out | awk '{if ($2 > 0.6) { print $0 }}' | wc -l >> $OUTFI

#
#   Rings: Fragment comparison and existing collection sets. Find number of systems
#   that are unique to comparison set (after property and substructural filtering)
#

echo "Calculating rings"
standardize -c "[*:1;R]-[*:2;R0]-[*:3;R]>>[*:1][*:2][*].[*][*:3]..[*:1;R0]-[*:2;R]-[*:3;R0]>>[*:1][*].[*][*:2][*].[*:3][*]..[*:1;R]-[*:2;R0]>>[*][*:1].[*][*:2]..[*:1;R0][*:2;R0]=[*:3;R]>>[*][*:1].[*][*:2]=[*:3]" -g -f smiles:u,T$NCCID $NCCROOT.std --log $NCCROOT.ring-err.log | grep -v SMILES | sed 's/\./\n/g' | awk '{print $1}' | grep [1] | sort -u > $NCCROOT.rng.smi
standardize -c "[*:1;R]-[*:2;R0]-[*:3;R]>>[*:1][*:2][*].[*][*:3]..[*:1;R0]-[*:2;R]-[*:3;R0]>>[*:1][*].[*][*:2][*].[*:3][*]..[*:1;R]-[*:2;R0]>>[*][*:1].[*][*:2]..[*:1;R0][*:2;R0]=[*:3;R]>>[*][*:1].[*][*:2]=[*:3]" -g -f smiles:u,T$COLID $COLROOT.std --log $COLROOT.ring-err.log | grep -v SMILES | sed 's/\./\n/g' | awk '{print $1}' | grep [1] | sort -u > $COLROOT.rng.smi 2> $NCCROOT.ring-err.log

comm -12 $NCCROOT.rng.smi $COLROOT.rng.smi > $NCCROOT.$COLROOT.rng.com.smi
echo "NCC Ring Scores: #Matched and #Comparison Set total (post-filtering)" >> $OUTFI
wc -l $NCCROOT.$COLROOT.rng.com.smi| awk '{print $1}' >> $OUTFI
wc -l $NCCROOT.rng.smi | awk '{print $1}' >> $OUTFI

#
#   Bemis-Murcko Frameworks: Calculate and count number that are unique to
#   comparison set (after property and substructural filtering)
#

echo "Calculating frameworks"
cxcalc frameworks -t bmf -s true $NCCROOT.std 2> $NCCROOT.frame-err.log | molconvert smiles:q -T $NCCID -Y -g | awk '{print $1}' | sort -u > $NCCROOT.bmf.smi
cxcalc frameworks -t bmf -s true $COLROOT.std 2> $COLROOT.frame-err.log | molconvert smiles:q -T $COLID -Y -g | awk '{print $1}' | sort -u > $COLROOT.bmf.smi

comm -12 $NCCROOT.bmf.smi $COLROOT.bmf.smi > $NCCROOT.$COLROOT.bmf.com.smi
echo "NCC Framework Scores: #Matched and #Comparison Set total (post-filtering)" >> $OUTFI
wc -l $NCCROOT.$COLROOT.bmf.com.smi | awk '{print $1}' >> $OUTFI
wc -l $NCCROOT.bmf.smi | awk '{print $1}' >> $OUTFI

echo "Analysis complete - Data summary in $OUTFI"
